import asyncio
import os.path
import posixpath
import tempfile
import unittest
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
from unittest import mock

from honesty.cache import Cache


class AiohttpStreamMock:
    def __init__(self, content: bytes) -> None:
        self._content = content

    # TODO async iterable[bytes]
    async def iter_any(self) -> Any:
        yield self._content


class AiohttpResponseMock:
    def __init__(self, content: bytes) -> None:
        self.content = AiohttpStreamMock(content)

    async def __aenter__(self) -> "AiohttpResponseMock":
        return self

    async def __aexit__(self, *args: Any) -> None:
        pass


class FakeCache:
    def __init__(
        self, path: str, url_to_contents: Dict[Tuple[str, Optional[str]], bytes]
    ) -> None:
        self.path: Path = Path(path)
        self.url_to_contents = url_to_contents
        self.json_index_url = "https://pypi.org/simple/"

    async def async_fetch(self, package_name: str, url: Optional[str] = None) -> Path:
        basename = posixpath.basename(url) if url else f"{package_name}_index.html"
        with open(self.path / basename, "wb") as f:
            f.write(self.url_to_contents[(package_name, url)])

        return self.path / basename


class CacheTest(unittest.TestCase):
    def test_fetch_caches(self) -> None:

        d = tempfile.mkdtemp()

        def get_side_effect(
            url: str, raise_for_status: bool = False, timeout: Any = None
        ) -> AiohttpResponseMock:
            if url == "https://example.com/other":
                return AiohttpResponseMock(b"other")
            elif url == "https://pypi.org/a/relpath":
                return AiohttpResponseMock(b"relpath")
            elif url == "https://pypi.org/simple/projectname/":
                return AiohttpResponseMock(b"foo")

            raise NotImplementedError(url)  # pragma: no cover

        with Cache(index_url="https://pypi.org/simple/", cache_dir=d) as cache:

            with mock.patch.object(cache.session, "get", side_effect=get_side_effect):
                rv = cache.fetch("projectname", url=None)
                self.assertTrue(rv.exists(), rv)
                self.assertEqual(
                    os.path.join(d, "pr", "oj", "projectname", "index.html"), str(rv)
                )
                rv = cache.fetch("projectname", url=None)
                self.assertEqual(
                    os.path.join(d, "pr", "oj", "projectname", "index.html"), str(rv)
                )
                # TODO mock_get.assert_called_once()
                with rv.open() as f:
                    self.assertEqual("foo", f.read())

                # Absolute path url support
                rv = cache.fetch("projectname", url="https://example.com/other")
                with rv.open() as f:
                    self.assertEqual("other", f.read())

                # Relative path support
                rv = cache.fetch("projectname", url="../../a/relpath")
                with rv.open() as f:
                    self.assertEqual("relpath", f.read())

    def test_cache_defaults(self) -> None:
        with Cache() as cache:
            self.assertEqual(
                Path("~/.cache/honesty/pypi").expanduser(), cache.cache_path
            )
            self.assertEqual("https://pypi.org/simple/", cache.index_url)

    @mock.patch("honesty.cache.os.environ.get")
    def test_cache_env_vars(self, mock_get: Any) -> None:
        mock_get.side_effect = {
            "HONESTY_CACHE": "/tmp",
            "HONESTY_INDEX_URL": "https://example.com/foo",
        }.get
        with Cache() as cache:
            self.assertEqual(Path("/tmp"), cache.cache_path)
            self.assertEqual("https://example.com/foo/", cache.index_url)

    def test_cache_invalid(self) -> None:
        with Cache() as cache:
            with self.assertRaises(NotImplementedError):
                # I no longer remember which project triggers this; in theory
                # all non-[a-z0-9-] should have been canonicalized away already.
                cache.fetch("pb&amp;j", url=None)

    def test_aenter(self) -> None:
        async def inner() -> None:
            async with Cache() as cache:
                self.assertTrue(cache)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(inner())

    def test_is_index(self) -> None:
        with Cache() as cache:
            self.assertTrue(cache._is_index_filename(None))
            self.assertTrue(cache._is_index_filename("json"))
            self.assertFalse(cache._is_index_filename("foo-0.1.tar.gz"))



import asyncio
import enum
import json
import re
import urllib.parse
from dataclasses import dataclass
from datetime import datetime, timezone
from html.parser import HTMLParser
from typing import Any, Dict, List, Optional, Tuple

from .cache import Cache

# Apologies in advance, "parsing" html via regex
CHECKSUM_RE = re.compile(
    r'\A(?P<url>[^"#]+\/(?P<basename>[^#]+))#(?P<checksum>[^="]+=[a-f0-9]+)\Z'
)
NUMERIC_VERSION = re.compile(
    r"^(?P<package>.*?)-(?P<version>[0-9][^-]*?)"
    r"(?P<suffix>(?P<platform>\.macosx|\.linux|\.cygwin|\.win(?:32|xp|))?-.*)?$"
)

ISO8601_FORMAT = "%Y-%m-%dT%H:%M:%S"


SDIST_EXTENSIONS = (".tar.gz", ".zip", ".tar.bz2")


# This list matches warehouse/packaging/models.py with the addition of UNKNOWN.
#
# Platform (in the case of bdist_dumb) is not currently stored anywhere but
# doesn't belong in this enum.

# Some rough popularity numbers at the time of writing (top 5k packages, with
# some double counts for the current release of each):
#
#     31 "bdist_dmg"
#     41 "bdist_rpm"
#    176 "bdist_msi"
#    309 "bdist_dumb"
#   6909 "bdist_wininst"
#  18201 "bdist_egg"
# 138984 "sdist"
# 155904 "bdist_wheel"


class FileType(enum.IntEnum):
    UNKNOWN = 0
    SDIST = 1  # .tar.gz or .zip (or for packages like Twisted, .tar.bz2)
    BDIST_DMG = 2  # .dmg
    BDIST_DUMB = 3  # -(platform).tar.gz
    BDIST_EGG = 4  # .egg
    BDIST_MSI = 5  # .msi
    BDIST_RPM = 6  # .rpm
    BDIST_WHEEL = 7  # .whl
    BDIST_WININST = 8  # .exe


class UnexpectedFilename(Exception):
    pass


def guess_file_type(filename: str) -> FileType:
    if filename.endswith(".egg"):
        return FileType.BDIST_EGG
    elif filename.endswith(".whl"):
        return FileType.BDIST_WHEEL
    elif filename.endswith(".exe"):
        return FileType.BDIST_WININST
    elif filename.endswith(".msi"):
        return FileType.BDIST_MSI
    elif filename.endswith(".rpm"):
        return FileType.BDIST_RPM
    elif filename.endswith(".dmg"):
        return FileType.BDIST_DMG
    elif filename.endswith(SDIST_EXTENSIONS):
        filename = remove_suffix(filename)
        match = NUMERIC_VERSION.match(filename)
        # Some oddly-named files are not likely to be loaded by pip either.
        if match is None:
            raise UnexpectedFilename(filename)
        # bdist_dumb can't be easily discerned
        if match.group("platform"):
            return FileType.BDIST_DUMB
        elif match.group("suffix") and match.group("suffix").startswith("-macosx"):
            return FileType.BDIST_DUMB
        return FileType.SDIST
    else:
        return FileType.UNKNOWN


@dataclass
class FileEntry:
    url: str  # https://files.pythonhosted.../foo-1.0.tgz
    basename: str  # foo-1.0.tgz
    checksum: str  # 'sha256=<foo>'
    file_type: FileType
    version: str  # TODO: better type
    requires_python: Optional[str] = None  # '>=3.6'
    python_version: Optional[str] = None  # 'py2.py3' or 'source'
    upload_time: Optional[datetime] = None
    # TODO extract upload date?

    @classmethod
    def from_attrs(cls, attrs: List[Tuple[str, Optional[str]]]) -> "FileEntry":
        """
        Given the <a> element's attrs from parsing the simple html index,
        returns a new FileEntry.
        """
        d = dict(attrs)
        if d["href"] is None:  # pragma: no cover
            raise KeyError("Empty href")
        m = CHECKSUM_RE.match(d["href"])
        if m is None:
            raise UnexpectedFilename(d["href"])
        url = m.group("url")
        basename = m.group("basename")
        checksum = m.group("checksum")

        return cls(
            url=url,
            basename=basename,
            checksum=checksum,
            file_type=guess_file_type(basename),
            version=guess_version(basename)[1],
            requires_python=d.get("data-requires-python"),
        )

    @classmethod
    def from_json(cls, version: str, obj: Dict[str, Any]) -> "FileEntry":
        # We still guess file_type here because warehouse gets it wrong for
        # bsdist_dumb and reports them as sdist.
        return cls(
            url=obj["url"],
            basename=obj["filename"],
            checksum=f"sha256={obj['digests']['sha256']}",
            file_type=guess_file_type(obj["filename"]),
            version=version,
            requires_python=obj["requires_python"],
            upload_time=parse_time(obj["upload_time_iso_8601"]),
        )


def parse_time(t: str) -> datetime:
    """Returns a parsed time with optional fractional seconds."""
    # Timestamps before ~2009-02-16 do not have fractional seconds.
    t = t.rstrip("Z")
    fmt, _, fractional = t.partition(".")

    # This makes it microseconds
    fractional = fractional[:6].ljust(6, "0")

    return datetime.strptime(t.split(".")[0], ISO8601_FORMAT).replace(
        microsecond=int(fractional), tzinfo=timezone.utc
    )


@dataclass
class PackageRelease:
    version: str
    files: List[FileEntry]


@dataclass
class Package:
    name: str
    releases: Dict[str, PackageRelease]


def remove_suffix(basename: str) -> str:
    suffixes = [
        ".egg",
        ".whl",
        ".zip",
        ".gz",
        ".bz2",
        ".tar",
        ".exe",
        ".msi",
        ".rpm",
        ".dmg",
    ]
    for s in suffixes:
        if basename.endswith(s):
            basename = basename[: -len(s)]
    return basename


# TODO itu-r-468-weighting-1.0.3.tar.gz
# TODO uttt-0.3-1.tar.gz
def guess_version(basename: str) -> Tuple[str, str]:
    """
    Returns (package name, version) or raises.
    """
    # This should use whatever setuptools/pip/etc use, but I spent about 10
    # minutes and couldn't find it tonight.
    basename = remove_suffix(basename)

    match = NUMERIC_VERSION.match(basename)
    if not match:
        raise UnexpectedFilename(basename)
    return match.group(1), match.group(2)


class LinkGatherer(HTMLParser):
    def __init__(self, strict: bool = False):
        super().__init__()
        self.entries: List[FileEntry] = []
        self.strict = strict

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
        if tag == "a":
            try:
                fe = FileEntry.from_attrs(attrs)
            except UnexpectedFilename:
                if not self.strict:
                    return
                raise

            self.entries.append(fe)


def parse_index(
    pkg: str, cache: Cache, strict: bool = False, use_json: bool = False
) -> Package:
    loop = asyncio.get_event_loop()
    package: Package = loop.run_until_complete(
        async_parse_index(pkg, cache, strict, use_json)
    )
    return package


async def async_parse_index(
    pkg: str, cache: Cache, strict: bool = False, use_json: bool = False
) -> Package:
    package = Package(name=pkg, releases={})
    if not use_json:
        # TODO: This preserves the input order, which is not based on proper
        # version comparisons.
        with open(await cache.async_fetch(pkg, url=None)) as f:
            gatherer = LinkGatherer(strict)
            gatherer.feed(f.read())

        for fe in gatherer.entries:
            v = fe.version
            if v not in package.releases:
                package.releases[v] = PackageRelease(version=v, files=[])
            package.releases[v].files.append(fe)
    else:
        # This will redirect away from canonical name if they differ
        # TODO: This doesn't obey environment variable, which we could
        url = urllib.parse.urljoin(cache.json_index_url, f"../pypi/{pkg}/json")
        with open(await cache.async_fetch(pkg, url=url)) as f:
            obj = json.loads(f.read())

        for k, release in obj["releases"].items():
            package.releases[k] = PackageRelease(version=k, files=[])
            for release_file in release:
                try:
                    package.releases[k].files.append(
                        FileEntry.from_json(k, release_file)
                    )
                except UnexpectedFilename:
                    if strict:
                        raise

    return package



"""
Cache-related stuff.
"""

import asyncio
import os
import posixpath
import urllib.parse
from pathlib import Path
from typing import Any, Dict, Optional

import aiohttp
import appdirs


def cache_dir(pkg: str) -> Path:
    a = pkg[:2]
    b = pkg[2:4] or "--"
    return Path(a, b, pkg)


DEFAULT_CACHE_DIR = os.path.join(
    appdirs.user_cache_dir("honesty", "python-packaging"), "pypi",
)
DEFAULT_HONESTY_INDEX_URL = "https://pypi.org/simple/"
BUFFER_SIZE = 4096 * 1024  # 4M


class Cache:
    def __init__(
        self,
        cache_dir: Optional[str] = None,
        index_url: Optional[str] = None,
        json_index_url: Optional[str] = None,
        fresh_index: bool = False,
        aiohttp_client_session_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not cache_dir:
            cache_dir = os.environ.get("HONESTY_CACHE", DEFAULT_CACHE_DIR)
        assert isinstance(cache_dir, str), cache_dir
        self.cache_path = Path(cache_dir).expanduser()

        if not index_url:
            index_url = os.environ.get("HONESTY_INDEX_URL", DEFAULT_HONESTY_INDEX_URL)
        assert isinstance(index_url, str), index_url
        if not index_url.endswith("/"):
            # in a browser, this would be a redirect; we don't know that here.
            index_url += "/"
        self.index_url = index_url

        if not json_index_url:
            json_index_url = os.environ.get("HONESTY_JSON_INDEX_URL", self.index_url)
        assert isinstance(json_index_url, str), json_index_url
        if not json_index_url.endswith("/"):
            index_url += "/"
        self.json_index_url = index_url

        self.fresh_index = fresh_index
        cskwargs: Dict[str, Any] = {
            "trust_env": True,
            "raise_for_status": True,
        }
        if aiohttp_client_session_kwargs is not None:
            cskwargs.update(aiohttp_client_session_kwargs)
        self.session = aiohttp.ClientSession(**cskwargs)

    def fetch(self, pkg: str, url: Optional[str]) -> Path:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.async_fetch(pkg, url))

    async def async_fetch(self, pkg: str, url: Optional[str]) -> Path:
        """
        When url=None, download the index.
        Otherwise, download (presumably) an archive.  url may be relative, and
        is presumably relative to the package index page.

        When self.fresh_index, never trust the cache for index (but still save).

        Returns a Path for where the cache wanted to save it.  We make effort to
        be concurrent-safe (last one wins).
        """

        # Because parse_index doesn't understand entities, there are some urls
        # that we currently get that we shouldn't bother fetching.
        if "&" in pkg or "#" in pkg:
            raise NotImplementedError("parse_index does not handle entities yet")

        pkg_url = urllib.parse.urljoin(self.index_url, f"{pkg}/")
        if url is None:
            url = pkg_url
        else:
            # pypi simple gives full urls, but if your mirror gives relative ones,
            # it's relative to the package's index page (which has trailing slash)
            url = urllib.parse.urljoin(pkg_url, url)

        filename = posixpath.basename(url)

        output_dir = self.cache_path / cache_dir(pkg)
        output_dir.mkdir(parents=True, exist_ok=True)

        output_file = output_dir / (filename or "index.html")

        if not output_file.exists() or (
            self.fresh_index and self._is_index_filename(filename)
        ):
            async with self.session.get(
                url, raise_for_status=True, timeout=None
            ) as resp:
                tmp = f"{output_file}.{os.getpid()}"
                with open(tmp, "wb") as f:
                    async for chunk in resp.content.iter_any():
                        f.write(chunk)
                # Last-writer-wins semantics
                os.rename(tmp, output_file)

        return output_file

    def _is_index_filename(self, name: Optional[str]) -> bool:
        return name in (None, "json")

    def __enter__(self) -> "Cache":
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> Any:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.session.close())

    async def __aenter__(self) -> "Cache":
        return self

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> Any:
        await self.session.close()



import datetime
import re
import tempfile
import unittest

from honesty.releases import (
    FileType,
    UnexpectedFilename,
    guess_file_type,
    guess_version,
    parse_index,
    parse_time,
)
from honesty.tests.cache import FakeCache

WOAH_INDEX_CONTENTS = b"""\
<!DOCTYPE html>
<html>
  <head>
    <title>Links for woah</title>
  </head>
  <body>
    <h1>Links for woah</h1>
    <a href="https://files.pythonhosted.org/packages/69/c9/a9951fcb2e706dd14cfc5d57a33eadc38a2b7477c82c12c229de5f6115db/woah-0.1-py3-none-any.whl#sha256=e705573ea8a88ec772174deea6a80c79f1e8b7e96130e27eee14b21d63f4e7f8" data-requires-python="&gt;=3.6">woah-0.1-py3-none-any.whl</a><br/>
    <a href="https://files.pythonhosted.org/packages/8f/3f/cd6d2edb9cf7049788db971fb5359cbde9fb28801d55b1aafa8f0df4813a/woah-0.1.tar.gz#sha256=d0760a3696271db53c361c950d93ceca7a022b5d739c0005e3bfb65785dd9d97" data-requires-python="&gt;=3.6">woah-0.1.tar.gz</a><br/>
    <a href="https://files.pythonhosted.org/packages/5e/95/871090fc9c10630d457b44967c9bb9c544b858cd3a2fe6dd60f9e169d99f/woah-0.2-py3-none-any.whl#sha256=e701a8d020a09fa32199cc74b386a3bf9730910fd46a6301fbb8203f287b27d7" data-requires-python="&gt;=3.6">woah-0.2-py3-none-any.whl</a><br/>
    <a href="https://files.pythonhosted.org/packages/fb/f2/dc6873f2763ffb457d3dbe4224ea59b21a8495fa0ef86d230b78cdba0f22/woah-0.2.tar.gz#sha256=62a886ed5e16506c039216dc0b5f342e72228e2038c750a1a7574321af6d8d68" data-requires-python="&gt;=3.6">woah-0.2.tar.gz</a><br/>
    </body>
</html>
<!--SERIAL 5860225-->
"""

WOAH_JSON_CONTENTS = b"""\
{"info":{"author":"Tim Hatch","author_email":"tim@timhatch.com","bugtrack_url":n
ull,"classifiers":["Development Status :: 4 - Beta","Environment :: Console","Li
cense :: OSI Approved :: Apache Software License","Programming Language :: Pytho
n","Programming Language :: Python :: 3","Programming Language :: Python :: 3.6"
,"Programming Language :: Python :: 3.7","Programming Language :: Python :: 3 ::
 Only","Topic :: Utilities"],"description":"# Woah\\n\\nWoah is a script that yo
u can wrap commands with, and it will wait until the\\nload average is reasonabl
e before running them.\\n\\n# Usage\\n\\nSimilar to `nice`, you just start your \
command with `woah` and everything after\\nthat will be run unchanged.\\n\\nSay,
 if you have a lot of other things going on, and want your backup to wait\\nfor
 things to settle down...\\n\\n```\\nwoah tar -xvzf /tmp/foo.tar users/tim\\n```\\
n\\nOr you\'re running multiple things with xargs, but some are more expensive
 than\\nothers and you want to keep your machine somewhat responsive...\\n\\n```\\
n(cd /users; ls -d *) | xargs -P32 -n1 --no-run-if-empty woah tar /tmp/{} /user
s/{}\\n```\\n\\n# Bugs and such\\n\\nhttps://github.com/thatch/woah\\n\\n# Licen
se\\n\\nApache 2.0\\n\\n\\n","description_content_type":"text/markdown","docs_ur
l":null,"download_url":"","downloads":{"last_day":-1,"last_month":-1,"last_week"
:-1},"home_page":"https://github.com/thatch/woah","keywords":"","license":"Apach
e 2.0","maintainer":"","maintainer_email":"","name":"woah","package_url":"https:
//pypi.org/project/woah/","platform":"","project_url":"https://pypi.org/project/
woah/","project_urls":{"Homepage":"https://github.com/thatch/woah"},"release_url
":"https://pypi.org/project/woah/0.2/","requires_dist":null,"requires_python":">
=3.6","summary":"Wait for reasonable load average","version":"0.2"},"last_serial
":5860225,"releases":{"0.1":[{"comment_text":"","digests":{"md5":"458804f2290028
8b7078ca1e3f39eb90","sha256":"e705573ea8a88ec772174deea6a80c79f1e8b7e96130e27eee
14b21d63f4e7f8"},"downloads":-1,"filename":"woah-0.1-py3-none-any.whl","has_sig"
:false,"md5_digest":"458804f22900288b7078ca1e3f39eb90","packagetype":"bdist_whee
l","python_version":"py3","requires_python":">=3.6","size":6511,"upload_time":"2
019-09-19T14:32:17","upload_time_iso_8601":"2019-09-19T14:32:17.358350Z","url":"
https://files.pythonhosted.org/packages/69/c9/a9951fcb2e706dd14cfc5d57a33eadc38a
2b7477c82c12c229de5f6115db/woah-0.1-py3-none-any.whl"},{"comment_text":"","diges
ts":{"md5":"8361f4eb6f0b5478540b39f851793b6b","sha256":"d0760a3696271db53c361c95
0d93ceca7a022b5d739c0005e3bfb65785dd9d97"},"downloads":-1,"filename":"woah-0.1.t
ar.gz","has_sig":false,"md5_digest":"8361f4eb6f0b5478540b39f851793b6b","packaget
ype":"sdist","python_version":"source","requires_python":">=3.6","size":1868,"up
load_time":"2019-09-19T14:32:19","upload_time_iso_8601":"2019-09-19T14:32:19.568
441Z","url":"https://files.pythonhosted.org/packages/8f/3f/cd6d2edb9cf7049788db9
71fb5359cbde9fb28801d55b1aafa8f0df4813a/woah-0.1.tar.gz"}],"0.2":[{"comment_text
":"","digests":{"md5":"79e7fd3d30a012b751d66ce881ccdc2a","sha256":"e701a8d020a09
fa32199cc74b386a3bf9730910fd46a6301fbb8203f287b27d7"},"downloads":-1,"filename":
"woah-0.2-py3-none-any.whl","has_sig":false,"md5_digest":"79e7fd3d30a012b751d66c
e881ccdc2a","packagetype":"bdist_wheel","python_version":"py3","requires_python"
:">=3.6","size":8355,"upload_time":"2019-09-20T05:39:40","upload_time_iso_8601":
"2019-09-20T05:39:40.581697Z","url":"https://files.pythonhosted.org/packages/5e/
95/871090fc9c10630d457b44967c9bb9c544b858cd3a2fe6dd60f9e169d99f/woah-0.2-py3-non
e-any.whl"},{"comment_text":"","digests":{"md5":"0b5eecde7203c8ff2260a51825dd1a9
c","sha256":"62a886ed5e16506c039216dc0b5f342e72228e2038c750a1a7574321af6d8d68"},
"downloads":-1,"filename":"woah-0.2.tar.gz","has_sig":false,"md5_digest":"0b5eec
de7203c8ff2260a51825dd1a9c","packagetype":"sdist","python_version":"source","req
uires_python":">=3.6","size":3255,"upload_time":"2019-09-20T05:39:41","upload_ti
me_iso_8601":"2019-09-20T05:39:41.862688Z","url":"https://files.pythonhosted.org
/packages/fb/f2/dc6873f2763ffb457d3dbe4224ea59b21a8495fa0ef86d230b78cdba0f22/woa
h-0.2.tar.gz"}]},"urls":[{"comment_text":"","digests":{"md5":"79e7fd3d30a012b751
d66ce881ccdc2a","sha256":"e701a8d020a09fa32199cc74b386a3bf9730910fd46a6301fbb820
3f287b27d7"},"downloads":-1,"filename":"woah-0.2-py3-none-any.whl","has_sig":fal
se,"md5_digest":"79e7fd3d30a012b751d66ce881ccdc2a","packagetype":"bdist_wheel","
python_version":"py3","requires_python":">=3.6","size":8355,"upload_time":"2019-
09-20T05:39:40","upload_time_iso_8601":"2019-09-20T05:39:40.581697Z","url":"http
s://files.pythonhosted.org/packages/5e/95/871090fc9c10630d457b44967c9bb9c544b858
cd3a2fe6dd60f9e169d99f/woah-0.2-py3-none-any.whl"},{"comment_text":"","digests":
{"md5":"0b5eecde7203c8ff2260a51825dd1a9c","sha256":"62a886ed5e16506c039216dc0b5f
342e72228e2038c750a1a7574321af6d8d68"},"downloads":-1,"filename":"woah-0.2.tar.g
z","has_sig":false,"md5_digest":"0b5eecde7203c8ff2260a51825dd1a9c","packagetype"
:"sdist","python_version":"source","requires_python":">=3.6","size":3255,"upload
_time":"2019-09-20T05:39:41","upload_time_iso_8601":"2019-09-20T05:39:41.862688Z
","url":"https://files.pythonhosted.org/packages/fb/f2/dc6873f2763ffb457d3dbe422
4ea59b21a8495fa0ef86d230b78cdba0f22/woah-0.2.tar.gz"}]}
""".replace(
    b"\n", b""
)

LONG_NAME = "scipy-0.14.1rc1.dev_205726a-cp33-cp33m-macosx_10_6_intel.macosx_10_9_intel.macosx_10_9_x86_64.macosx_10_10_intel.macosx_10_10_x86_64.whl"


class ReleasesTest(unittest.TestCase):
    def test_get_entries(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            c = FakeCache(d, {("woah", None): WOAH_INDEX_CONTENTS})
            pkg = parse_index("woah", c)  # type: ignore

        self.assertEqual("woah", pkg.name)
        self.assertEqual(2, len(pkg.releases))

        v01 = pkg.releases["0.1"]
        self.assertEqual(2, len(v01.files))

        self.assertEqual(
            "https://files.pythonhosted.org/packages/69/c9/a9951fcb2e706dd14cfc5d57a33eadc38a2b7477c82c12c229de5f6115db/woah-0.1-py3-none-any.whl",
            v01.files[0].url,
        )
        self.assertEqual("woah-0.1-py3-none-any.whl", v01.files[0].basename)
        self.assertEqual(
            "sha256=e705573ea8a88ec772174deea6a80c79f1e8b7e96130e27eee14b21d63f4e7f8",
            v01.files[0].checksum,
        )
        self.assertEqual(">=3.6", v01.files[0].requires_python)
        self.assertEqual(None, v01.files[0].upload_time)

    def test_get_entries_json(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            c = FakeCache(
                d, {("woah", "https://pypi.org/pypi/woah/json"): WOAH_JSON_CONTENTS}
            )
            pkg = parse_index("woah", c, use_json=True)  # type: ignore

        self.assertEqual("woah", pkg.name)
        self.assertEqual(2, len(pkg.releases))

        v01 = pkg.releases["0.1"]
        self.assertEqual(2, len(v01.files))

        self.assertEqual(
            "https://files.pythonhosted.org/packages/69/c9/a9951fcb2e706dd14cfc5d57a33eadc38a2b7477c82c12c229de5f6115db/woah-0.1-py3-none-any.whl",
            v01.files[0].url,
        )
        self.assertEqual("woah-0.1-py3-none-any.whl", v01.files[0].basename)
        self.assertEqual(
            "sha256=e705573ea8a88ec772174deea6a80c79f1e8b7e96130e27eee14b21d63f4e7f8",
            v01.files[0].checksum,
        )
        self.assertEqual(">=3.6", v01.files[0].requires_python)
        self.assertEqual(
            datetime.datetime(
                2019, 9, 19, 14, 32, 17, 358350, tzinfo=datetime.timezone.utc
            ),
            v01.files[0].upload_time,
        )

    def test_error_on_unexpected_filename_regex(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            c = FakeCache(
                d, {("woah", None): re.sub(rb'#.*?"', b'"', WOAH_INDEX_CONTENTS)}
            )
            with self.assertRaises(UnexpectedFilename):
                parse_index("woah", c, strict=True)  # type: ignore

    def test_strict(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            c = FakeCache(
                d, {("woah", None): WOAH_INDEX_CONTENTS.replace(b"woah-0.1", b"woah")}
            )
            with self.assertRaises(UnexpectedFilename):
                parse_index("woah", c, strict=True)  # type: ignore

    def test_non_strict(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            c = FakeCache(
                d, {("woah", None): WOAH_INDEX_CONTENTS.replace(b"woah-0.1", b"woah")}
            )
            pkg = parse_index("woah", c, strict=False)  # type: ignore

        self.assertEqual(1, len(pkg.releases))

        v02 = pkg.releases["0.2"]
        self.assertEqual(2, len(v02.files))

    def test_guess_version(self) -> None:
        self.assertEqual(("foo", "0.1"), guess_version("foo-0.1.tar.gz"))
        self.assertEqual(("foo", "0.1"), guess_version("foo-0.1-py3-none.whl"))
        self.assertEqual(("foo", "0.1"), guess_version("foo-0.1-any-none.whl"))
        with self.assertRaises(UnexpectedFilename):
            guess_version("foo.tar.gz")

        self.assertEqual(("scipy", "0.14.1rc1.dev_205726a"), guess_version(LONG_NAME))
        self.assertEqual(
            ("javatools", "1.4.0"),
            guess_version("javatools-1.4.0.macosx-10.14-x86_64.tar.gz"),
        )
        self.assertEqual(("pypi", "2"), guess_version("pypi-2.tar.gz"))
        self.assertEqual(
            ("psutil", "5.3.0"), guess_version("psutil-5.3.0.win-amd64-py3.6.exe"),
        )
        self.assertEqual(
            ("psutil", "5.3.0"), guess_version("psutil-5.3.0.win32-py3.6.exe"),
        )

    def test_guess_file_type(self) -> None:
        expected = [
            ("foo-0.1", FileType.UNKNOWN),
            ("foo-0.1.tar.gz", FileType.SDIST),
            ("pypi-2.tar.gz", FileType.SDIST),
            ("foo-0.1.dmg", FileType.BDIST_DMG),
            # These two are real examples, and yes there's a dot vs dash
            # discrepancy
            ("javatools-1.4.0.macosx-10.14-x86_64.tar.gz", FileType.BDIST_DUMB),
            ("pyre-check-0.0.29-macosx_10_11_x86_64.tar.gz", FileType.BDIST_DUMB),
            ("foo-0.1.egg", FileType.BDIST_EGG),
            ("foo-0.1.msi", FileType.BDIST_MSI),
            ("foo-0.1.rpm", FileType.BDIST_RPM),
            ("foo-0.1-manylinux.whl", FileType.BDIST_WHEEL),
            ("foo-0.1.exe", FileType.BDIST_WININST),
        ]

        for a, b in expected:
            with self.subTest(a):
                self.assertEqual(b, guess_file_type(a))

        with self.assertRaises(UnexpectedFilename):
            guess_file_type("ibm_db.tar.gz")

    def test_parse_time(self) -> None:
        v = parse_time("2019-09-19T14:32:17.358350")
        self.assertEqual(
            datetime.datetime(
                2019, 9, 19, 14, 32, 17, 358350, tzinfo=datetime.timezone.utc
            ),
            v,
        )

        v = parse_time("2019-09-19T14:32:17")
        self.assertEqual(
            datetime.datetime(2019, 9, 19, 14, 32, 17, 0, tzinfo=datetime.timezone.utc),
            v,
        )

        v = parse_time("2019-09-19T14:32:17.123")
        self.assertEqual(
            datetime.datetime(
                2019, 9, 19, 14, 32, 17, 123000, tzinfo=datetime.timezone.utc
            ),
            v,
        )



from builtins import _PathLike
from typing import List, Optional, Any, Type, Union, Text, IO

class Archive:
    member_names: List[str]
    def member_is_file(self, name: str) -> bool: ...
    # TODO overload?
    def open_member(self, name: str, mode: str) -> IO[Any]: ...
    def extract(self, path: Any=..., members: Any=...) -> None: ...
    def __enter__(self) -> Archive: ...
    def __exit__(self, *args: Any) -> None: ...

class ZipArchive(Archive):
    pass

def open(name: Union[Text, '_PathLike[Any]'], mode: str, engine: Optional[Type[Archive]]=...) -> Archive: ...



import asyncio
import functools
import json
import os.path
import shutil
import sys
from datetime import datetime, timezone
from enum import Enum, IntEnum
from pathlib import Path
from typing import Any, List, Optional

import click
import pkg_resources

from honesty.__version__ import __version__
from honesty.api import async_download_many
from honesty.archive import extract_and_get_names
from honesty.cache import Cache
from honesty.checker import guess_license, has_nativemodules, is_pep517, run_checker
from honesty.releases import FileType, Package, async_parse_index, parse_index


# TODO type
def wrap_async(coro: Any) -> Any:
    @functools.wraps(coro)
    def inner(*args: Any, **kwargs: Any) -> Any:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(coro(*args, **kwargs))

    return inner


def dataclass_default(obj: Any) -> Any:
    if hasattr(obj, "__dataclass_fields__"):
        return obj.__dict__
    elif isinstance(obj, (Enum, IntEnum)):
        return obj.name
    elif isinstance(obj, datetime):
        return str(obj)
    else:
        raise TypeError(obj)


@click.group()
@click.version_option(__version__, prog_name="honesty")
def cli() -> None:
    pass


@cli.command(help="List available archives")
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.option("--as_json", is_flag=True, type=bool)
@click.argument("package_name")
@wrap_async
async def list(fresh: bool, nouse_json: bool, as_json: bool, package_name: str) -> None:
    async with Cache(fresh_index=fresh) as cache:
        package = await async_parse_index(package_name, cache, use_json=not nouse_json)

    if as_json:
        for k, v in package.releases.items():
            print(json.dumps(v, default=dataclass_default, sort_keys=True))
    else:
        print(f"package {package.name}")
        print("releases:")
        for k, v in package.releases.items():
            print(f"  {k}:")
            for f in v.files:
                if f.requires_python:
                    print(f"    {f.basename} (requires_python {f.requires_python})")
                else:
                    print(f"    {f.basename}")


@cli.command(help="Check for consistency among archives")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.argument("package_name")
def check(verbose: bool, fresh: bool, nouse_json: bool, package_name: str) -> None:
    with Cache(fresh_index=fresh) as cache:
        package_name, operator, version = package_name.partition("==")
        package = parse_index(package_name, cache, use_json=not nouse_json)
        selected_versions = select_versions(package, operator, version)

        if verbose:
            click.echo(f"check {package_name} {selected_versions}")

        rc = 0
        for v in selected_versions:
            rc |= run_checker(package, v, verbose=verbose, cache=cache)

    if rc != 0:
        sys.exit(rc)


@cli.command(help="Check for presence of pep517 markers")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.argument("package_name")
def ispep517(verbose: bool, fresh: bool, nouse_json: bool, package_name: str) -> None:
    with Cache(fresh_index=fresh) as cache:
        package_name, operator, version = package_name.partition("==")
        package = parse_index(package_name, cache, use_json=not nouse_json)
        selected_versions = select_versions(package, operator, version)

        if verbose:
            click.echo(f"check {package_name} {selected_versions}")

        rc = 0
        for v in selected_versions:
            rc |= is_pep517(package, v, verbose=verbose, cache=cache)

    if rc != 0:
        sys.exit(rc)


@cli.command(help="Check for native modules in bdist")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.argument("package_name")
def native(verbose: bool, fresh: bool, nouse_json: bool, package_name: str) -> None:
    with Cache(fresh_index=fresh) as cache:
        package_name, operator, version = package_name.partition("==")
        package = parse_index(package_name, cache, use_json=not nouse_json)
        selected_versions = select_versions(package, operator, version)

        if verbose:
            click.echo(f"check {package_name} {selected_versions}")

        rc = 0
        for v in selected_versions:
            rc |= has_nativemodules(package, v, verbose=verbose, cache=cache)

    if rc != 0:
        sys.exit(rc)


@cli.command(help="Guess license of a package")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.argument("package_name")
def license(verbose: bool, fresh: bool, nouse_json: bool, package_name: str) -> None:
    with Cache(fresh_index=fresh) as cache:
        package_name, operator, version = package_name.partition("==")
        package = parse_index(package_name, cache, use_json=not nouse_json)
        selected_versions = select_versions(package, operator, version)

        if verbose:
            click.echo(f"check {package_name} {selected_versions}")

        rc = 0
        for v in selected_versions:
            license = guess_license(package, v, verbose=verbose, cache=cache)
            if license is not None and not isinstance(license, str):
                license = license.shortname
            if license is None:
                rc |= 1
            print(f"{package_name}=={v}: {license or 'Unknown'}")

    if rc != 0:
        sys.exit(rc)


@cli.command(help="Download an sdist, print path on stdout")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.option("--dest", help="Directory to store in", default="")
@click.option(
    "--index-url", help="Alternate index url (uses HONESTY_INDEX_URL or pypi by default"
)
@click.argument("package_name")
@wrap_async
async def download(
    verbose: bool,
    fresh: bool,
    nouse_json: bool,
    dest: str,
    index_url: Optional[str],
    package_name: str,
) -> None:
    dest_path: Optional[Path]
    if dest:
        dest_path = Path(dest)
        dest_path.mkdir(parents=True, exist_ok=True)
    else:
        dest_path = None

    async with Cache(fresh_index=fresh, index_url=index_url) as cache:
        package_name, operator, version = package_name.partition("==")
        package = await async_parse_index(package_name, cache, use_json=not nouse_json)
        selected_versions = select_versions(package, operator, version)

        if verbose:
            click.echo(f"check {package_name} {selected_versions}")

        rc = await async_download_many(
            package, versions=selected_versions, dest=dest_path, cache=cache
        )

    sys.exit(rc)


@cli.command(help="Download/extract an sdist, print path on stdout")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--nouse_json", is_flag=True, type=bool)
@click.option("--dest", help="Directory to store in", default="")
@click.option(
    "--index-url", help="Alternate index url (uses HONESTY_INDEX_URL or pypi by default"
)
@click.argument("package_name")
@wrap_async
async def extract(
    verbose: bool,
    fresh: bool,
    nouse_json: bool,
    dest: str,
    index_url: Optional[str],
    package_name: str,
) -> None:

    async with Cache(fresh_index=fresh, index_url=index_url) as cache:
        package_name, operator, version = package_name.partition("==")
        package = await async_parse_index(package_name, cache, use_json=not nouse_json)
        selected_versions = select_versions(package, operator, version)
        if len(selected_versions) != 1:
            raise click.ClickException(f"Wrong number of versions: {selected_versions}")

        if verbose:
            click.echo(f"check {package_name} {selected_versions}")

        rel = package.releases[selected_versions[0]]
        sdists = [f for f in rel.files if f.file_type == FileType.SDIST]
        if not sdists:
            raise click.ClickException(f"{package.name} no sdists")

        lp = await cache.async_fetch(pkg=package_name, url=sdists[0].url)

        archive_root, _ = extract_and_get_names(
            lp, strip_top_level=True, patterns=("*.*",)
        )

        subdirs = tuple(Path(archive_root).iterdir())
        if dest:
            for subdir in subdirs:
                shutil.copytree(subdir, Path(dest, subdir.name))
        else:
            dest = archive_root

        # Try to be helpful in the common case that there's a top-level
        # directory by itself.  Specifying a non-empty dest makes the fallback
        # less useful.
        if len(subdirs) == 1:
            print(os.path.join(dest, subdirs[0].name))
        else:
            print(dest)


@cli.command(help="Print age in days for a given release")
@click.option("--verbose", "-v", is_flag=True, type=bool)
@click.option("--fresh", "-f", is_flag=True, type=bool)
@click.option("--base", help="yyyy-mm-dd of when to subtract from")
@click.argument("package_name")
@wrap_async
async def age(verbose: bool, fresh: bool, base: str, package_name: str,) -> None:

    if base:
        base_date = datetime.strptime(base, "%Y-%m-%d")
    else:
        base_date = datetime.utcnow()
    base_date = base_date.replace(tzinfo=timezone.utc)

    async with Cache(fresh_index=fresh) as cache:
        package_name, operator, version = package_name.partition("==")
        package = await async_parse_index(package_name, cache, use_json=True)
        selected_versions = select_versions(package, operator, version)
        for v in selected_versions:
            t = min(x.upload_time for x in package.releases[v].files)
            assert t is not None

            diff = base_date - t
            days = diff.days + (diff.seconds / 86400.0)
            print(f"{v}\t{t.strftime('%Y-%m-%d')}\t{days:.2f}")


def select_versions(package: Package, operator: str, selector: str) -> List[str]:
    """
    Given operator='==' and selector='*' or '2.0', return a list of the matching
    versions, in increasing order.
    """
    if not package.releases:
        raise click.ClickException(f"No releases at all for {package.name}")

    if operator not in ("", "=="):
        raise click.ClickException("Only '==' is supported")

    if selector == "":
        # latest
        version = sorted(package.releases, key=pkg_resources.parse_version)[-1]
        return [version]
    elif selector == "*":
        versions: List[str] = sorted(package.releases, key=pkg_resources.parse_version)
        return versions
    else:
        if selector not in package.releases:
            raise click.ClickException(
                f"The version {selector} does not exist for {package.name}"
            )
        return [selector]


if __name__ == "__main__":
    cli()



import os
import os.path
import shutil
import tempfile
import unittest
from pathlib import Path
from typing import Dict
from unittest import mock

from honesty.archive import archive_hashes, extract_and_get_names


def create_test_archive(
    path_contents: Dict[str, str], extension: str, format: str
) -> Path:
    """
    Create an archive with the specified characteristics.

    Caller is responsible for deleting.
    """
    with tempfile.TemporaryDirectory() as d:
        p = Path(d)
        for path, contents in path_contents.items():
            (p / path).parent.mkdir(parents=True, exist_ok=True)
            with (p / path).open("w") as f:
                f.write(contents)

        name = tempfile.mktemp(suffix=f".{extension}")
        short_name = name.rsplit(".", 1)[0]
        tmp = shutil.make_archive(short_name, format, d)
        os.rename(tmp, name)
        return Path(name)


class ArchiveTest(unittest.TestCase):
    def test_extract(self) -> None:
        archive = create_test_archive(
            {
                "foo-0.1/setup.py": "setup()\n",
                "foo-0.1/src/proj/__init__.py": "",
                "foo-0.1/pyproject.toml": "[section]\n",
            },
            "whl",
            "zip",
        )
        try:
            with tempfile.TemporaryDirectory() as d:
                with mock.patch("honesty.archive.os.environ.get", return_value=d):
                    archive_root, names = extract_and_get_names(archive)
                    self.assertEqual(2, len(names))
                    # We didn't specify strip_top_level so these are both as in
                    # archive.
                    self.assertEqual(
                        {
                            os.path.join("foo-0.1", "setup.py"),
                            os.path.join("foo-0.1", "src", "proj", "__init__.py"),
                        },
                        {n[1] for n in names},
                    )
                    self.assertEqual(
                        {
                            os.path.join("foo-0.1", "setup.py"),
                            os.path.join("foo-0.1", "src", "proj", "__init__.py"),
                        },
                        {n[0] for n in names},
                    )

                    # We can call it a second time  with different args, and it
                    # doesn't actually extract again.
                    archive_root, names = extract_and_get_names(
                        archive, strip_top_level=True
                    )
                    self.assertEqual(2, len(names))
                    self.assertEqual(
                        {"setup.py", os.path.join("proj", "__init__.py")},
                        {n[1] for n in names},
                    )
                    self.assertEqual(
                        {
                            os.path.join("foo-0.1", "setup.py"),
                            os.path.join("foo-0.1", "src", "proj", "__init__.py"),
                        },
                        {n[0] for n in names},
                    )

                    # Another potential edge case, when patterns change
                    # (currently we extract everything the first time)
                    archive_root, names = extract_and_get_names(
                        archive, patterns=("*.toml",)
                    )
                    self.assertEqual(1, len(names))
                    self.assertEqual(
                        (
                            os.path.join("foo-0.1", "pyproject.toml"),
                            os.path.join("foo-0.1", "pyproject.toml"),
                        ),
                        names[0],
                    )
        finally:
            os.remove(archive)

    def test_hashes(self) -> None:
        archive = create_test_archive(
            {
                "foo-0.1/setup.py": "setup()\n",
                "foo-0.1/src/proj/__init__.py": "",
                "foo-0.1/pyproject.toml": "[section]\n",
            },
            "whl",
            "zip",
        )
        try:
            with tempfile.TemporaryDirectory() as d:
                with mock.patch("honesty.archive.os.environ.get", return_value=d):
                    hashes = archive_hashes(archive)
                    self.assertEqual(
                        {
                            os.path.join(
                                "foo-0.1", "setup.py"
                            ): "f568932ab271783a0234a22ed902131b7dfef0a9",
                            os.path.join(
                                "foo-0.1", "src", "proj", "__init__.py"
                            ): "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        },
                        hashes,
                    )

        finally:
            os.remove(archive)



from setuptools import setup

setup(
    use_scm_version=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
    ],
    entry_points={"console_scripts": ["honesty = honesty.cmdline:cli"]},
    install_requires=[
        "aiohttp >= 3.6",
        "appdirs >= 1.4",
        "click >= 7.0",
        "dataclasses >= 0.7; python_version < '3.7'",
        "infer-license >= 0.0.6",
    ],
)
